/**
 * HTTP endpoint declaration.
 *
 * @since 1.0.0
 */
import type * as HttpMethod from "@effect/platform/HttpMethod";
import type * as HttpRouter from "@effect/platform/HttpRouter";
import type * as Array from "effect/Array";
import type * as Pipeable from "effect/Pipeable";
import type * as Schema from "effect/Schema";
import type * as Types from "effect/Types";
import type * as ApiRequest from "./ApiRequest.js";
import type * as ApiResponse from "./ApiResponse.js";
import type * as ApiSchema from "./ApiSchema.js";
import type * as Representation from "./Representation.js";
import type * as Security from "./Security.js";
/**
 * @since 1.0.0
 * @category type id
 */
export declare const TypeId: unique symbol;
/**
 * @since 1.0.0
 * @category type id
 */
export type TypeId = typeof TypeId;
/**
 * @category models
 * @since 1.0.0
 */
export interface Options {
    readonly deprecated?: boolean;
    readonly description?: string;
    readonly summary?: string;
}
/**
 * @category models
 * @since 1.0.0
 */
export interface ApiEndpoint<Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, _Security extends Security.Security.Any> extends ApiEndpoint.Variance<Id, Request, Response, _Security>, Pipeable.Pipeable {
}
/**
 * @category models
 * @since 1.0.0
 */
export declare namespace ApiEndpoint {
    /**
     * @category models
     * @since 1.0.0
     */
    interface Variance<Id, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any> {
        readonly [TypeId]: {
            readonly _Id: Types.Covariant<Id>;
            readonly _Request: Types.Covariant<Request>;
            readonly _Response: Types.Covariant<Response>;
            readonly _Security: Types.Covariant<Security>;
        };
    }
    /**
     * Any endpoint id.
     *
     * @category models
     * @since 1.0.0
     */
    type AnyId = string;
    /**
     * Any endpoint with `Request = Request.Any` and `Response = Response.Any`.
     *
     * @category models
     * @since 1.0.0
     */
    type Any = ApiEndpoint<AnyId, ApiRequest.ApiRequest.Any, ApiResponse.ApiResponse.Any, Security.Security.Any>;
    /**
     * Default endpoint spec.
     *
     * @category models
     * @since 1.0.0
     */
    type Default<Id extends AnyId> = ApiEndpoint<Id, ApiRequest.ApiRequest.Default, ApiResponse.ApiResponse.Default, Security.Security.Default>;
    /**
     * @category models
     * @since 1.0.0
     */
    type Context<Endpoint> = [Endpoint] extends [ApiEndpoint<any, infer Request, infer Response, infer S>] ? ApiRequest.ApiRequest.Context<Request> | ApiResponse.ApiResponse.Context<Response> | Security.Security.Context<S> : never;
    /**
     * @category models
     * @since 1.0.0
     */
    type ContextNoSecurity<Endpoint> = [Endpoint] extends [ApiEndpoint<any, infer Request, infer Response, any>] ? ApiRequest.ApiRequest.Context<Request> | ApiResponse.ApiResponse.Context<Response> : never;
    /**
     * @category models
     * @since 1.0.0
     */
    type ExtractById<Endpoint, Id extends AnyId> = Endpoint extends ApiEndpoint<Id, any, any, any> ? Endpoint : never;
    /**
     * @category models
     * @since 1.0.0
     */
    type ExcludeById<Endpoint extends Any, Id extends AnyId> = Endpoint extends ApiEndpoint<Id, any, any, any> ? never : Endpoint;
    /**
     * @category models
     * @since 1.0.0
     */
    type Id<Endpoint> = Endpoint extends ApiEndpoint<infer Id, any, any, any> ? Id : never;
    /**
     * @category models
     * @since 1.0.0
     */
    type Request<Endpoint> = Endpoint extends ApiEndpoint<any, infer Request, any, any> ? Request : never;
    /**
     * @category models
     * @since 1.0.0
     */
    type Response<Endpoint> = Endpoint extends ApiEndpoint<any, any, infer Response, any> ? Response : never;
    /**
     * @category models
     * @since 1.0.0
     */
    type Security<Endpoint> = Endpoint extends ApiEndpoint<any, any, any, infer Security> ? Security : never;
    /**
     * @category models
     * @since 1.0.0
     */
    type Error<Endpoint> = [Endpoint] extends [ApiEndpoint<any, any, any, infer S>] ? Security.Security.Error<S> : never;
}
/**
 * @category constructors
 * @since 1.0.0
 */
export declare const make: <Id extends ApiEndpoint.AnyId>(method: HttpMethod.HttpMethod, id: Id, path: HttpRouter.PathInput, options?: Partial<Options>) => ApiEndpoint.Default<Id>;
/**
 * @category constructors
 * @since 1.0.0
 */
export declare const get: <Id extends ApiEndpoint.AnyId>(id: Id, path: HttpRouter.PathInput, options?: Partial<Options>) => ApiEndpoint.Default<Id>;
/**
 * @category constructors
 * @since 1.0.0
 */
export declare const post: <Id extends ApiEndpoint.AnyId>(id: Id, path: HttpRouter.PathInput, options?: Partial<Options>) => ApiEndpoint.Default<Id>;
/**
 * @category constructors
 * @since 1.0.0
 */
export declare const put: <Id extends ApiEndpoint.AnyId>(id: Id, path: HttpRouter.PathInput, options?: Partial<Options>) => ApiEndpoint.Default<Id>;
/**
 * @category constructors
 * @since 1.0.0
 */
export declare const patch: <Id extends ApiEndpoint.AnyId>(id: Id, path: HttpRouter.PathInput, options?: Partial<Options>) => ApiEndpoint.Default<Id>;
declare const _delete: <Id extends ApiEndpoint.AnyId>(id: Id, path: HttpRouter.PathInput, options?: Partial<Options>) => ApiEndpoint.Default<Id>;
export { 
/**
 * @category constructors
 * @since 1.0.0
 */
_delete as delete };
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setRequest: <Request extends ApiRequest.ApiRequest.Any>(request: Request) => <Id extends ApiEndpoint.AnyId, _ extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, _, Response, Security>) => ApiEndpoint<Id, Request, Response, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setRequestBody: <B, R2>(schema: Schema.Schema<B, any, R2>) => <Id extends ApiEndpoint.AnyId, _, P, Q, H, R1, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, ApiRequest.ApiRequest<_, P, Q, H, R1>, Response, Security>) => ApiEndpoint<Id, ApiRequest.ApiRequest<B, P, Q, H, R1 | R2>, Response, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setRequestPath: <P, I extends Readonly<Record<string, string | undefined>>, R2>(schema: Schema.Schema<P, I, R2>) => <Id extends ApiEndpoint.AnyId, B, _, Q, H, R1, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, ApiRequest.ApiRequest<B, _, Q, H, R1>, Response, Security>) => ApiEndpoint<Id, ApiRequest.ApiRequest<B, P, Q, H, R1 | R2>, Response, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setRequestQuery: <Q, I extends Readonly<Record<string, string | undefined | ReadonlyArray<string>>>, R2>(schema: Schema.Schema<Q, I, R2>) => <Id extends ApiEndpoint.AnyId, B, P, _, H, R1, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, ApiRequest.ApiRequest<B, P, _, H, R1>, Response, Security>) => ApiEndpoint<Id, ApiRequest.ApiRequest<B, P, Q, H, R1 | R2>, Response, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setRequestHeaders: <H, I extends Readonly<Record<string, string | undefined>>, R2>(schema: Schema.Schema<H, I, R2>) => <Id extends ApiEndpoint.AnyId, B, P, Q, _, R1, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, ApiRequest.ApiRequest<B, P, Q, _, R1>, Response, Security>) => ApiEndpoint<Id, ApiRequest.ApiRequest<B, P, Q, H, R1 | R2>, Response, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setSecurity: <Security extends Security.Security.Any>(security: Security) => <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, _ extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, _>) => ApiEndpoint<Id, Request, Response, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setOptions: (options: Partial<Options>) => <A extends ApiEndpoint.Any>(endpoint: A) => A;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const getId: <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => Id;
/**
 * @category getters
 * @since 1.0.0
 */
export declare const getRequest: <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => Request;
/**
 * @category getters
 * @since 1.0.0
 */
export declare const getResponse: <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => ReadonlyArray<Response>;
/**
 * @category getters
 * @since 1.0.0
 */
export declare const getSecurity: <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => Security;
/**
 * @category getters
 * @since 1.0.0
 */
export declare const getPath: <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => HttpRouter.PathInput;
/**
 * @category getters
 * @since 1.0.0
 */
export declare const getMethod: <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => HttpMethod.HttpMethod;
/**
 * @category getters
 * @since 1.0.0
 */
export declare const getOptions: <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => Options;
/**
 * @category getters
 * @since 1.0.0
 */
export declare const setResponse: <Status extends ApiResponse.ApiResponse.AnyStatus, Body = ApiSchema.Ignored, Headers = ApiSchema.Ignored, R = never>(response: {
    readonly status: Status;
    readonly body?: Schema.Schema<Body, any, R>;
    readonly headers?: Schema.Schema<Headers, any, R>;
} | ApiResponse.ApiResponse<Status, Body, Headers, R>) => <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, _ extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, _, Security>) => ApiEndpoint<Id, Request, ApiResponse.ApiResponse<Status, Body, Headers, R>, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setResponseStatus: <Status extends ApiResponse.ApiResponse.AnyStatus>(status: Status) => <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, _ extends ApiResponse.ApiResponse.AnyStatus, B, H, R, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, ApiResponse.ApiResponse<_, B, H, R>, Security>) => ApiEndpoint<Id, Request, ApiResponse.ApiResponse<Status, B, H, R>, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setResponseBody: <B, R2>(schema: Schema.Schema<B, any, R2>) => <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, S extends ApiResponse.ApiResponse.AnyStatus, _, H, R1, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, ApiResponse.ApiResponse<S, _, H, R1>, Security>) => ApiEndpoint<Id, Request, ApiResponse.ApiResponse<S, B, H, R1 | R2>, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setResponseHeaders: <H, R2>(schema: Schema.Schema<H, any, R2>) => <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, S extends ApiResponse.ApiResponse.AnyStatus, B, _, R1, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, ApiResponse.ApiResponse<S, B, _, R1>, Security>) => ApiEndpoint<Id, Request, ApiResponse.ApiResponse<S, B, H, R1 | R2>, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const addResponse: <Status extends ApiResponse.ApiResponse.AnyStatus, Body = ApiSchema.Ignored, Headers = ApiSchema.Ignored, R = never>(response: ApiResponse.ApiResponse<Status, Body, Headers, R> | {
    readonly status: Status;
    readonly body?: Schema.Schema<Body, any, R>;
    readonly headers?: Schema.Schema<Headers, any, R>;
}) => <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response1 extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response1, Security>) => ApiEndpoint<Id, Request, Response1 | ApiResponse.ApiResponse<Status, Body, Headers, R>, Security>;
/**
 * @category modifications
 * @since 1.0.0
 */
export declare const setResponseRepresentations: (representations: Array.NonEmptyReadonlyArray<Representation.Representation>) => <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => ApiEndpoint<Id, Request, Response, Security>;
/**
 * Is used by the client and server implementation to determine whether
 * to use a full `{ status; body; headers }` response or just the body.
 *
 * The logic is that if there is only one declared response and it doesn't
 * specify any headers, the simplified version with body only is used, otherwise
 * the full response is used.
 *
 * @category getters
 * @since 1.0.0
 */
export declare const isFullResponse: <Id extends ApiEndpoint.AnyId, Request extends ApiRequest.ApiRequest.Any, Response extends ApiResponse.ApiResponse.Any, Security extends Security.Security.Any>(endpoint: ApiEndpoint<Id, Request, Response, Security>) => boolean;
/**
 * @category refinements
 * @since 1.0.0
 */
export declare const isApiEndpoint: (u: unknown) => u is ApiEndpoint.Any;
//# sourceMappingURL=ApiEndpoint.d.ts.map