import * as Array from "effect/Array";
import * as Equivalence from "effect/Equivalence";
import { pipe } from "effect/Function";
import * as Order from "effect/Order";
import * as Pipeable from "effect/Pipeable";
import * as Predicate from "effect/Predicate";
import * as Tuple from "effect/Tuple";
import * as ApiRequest from "../ApiRequest.js";
import * as ApiResponse from "../ApiResponse.js";
import * as ApiSchema from "../ApiSchema.js";
import * as Security from "../Security.js";
import * as api_request from "./api-request.js";
import * as api_response from "./api-response.js";
/** @internal */
export const TypeId = /*#__PURE__*/Symbol.for("effect-http/Api/EndpointTypeId");
/** @internal */
export const variance = {
  /* c8 ignore next */
  _Id: _ => _,
  /* c8 ignore next */
  _Request: _ => _,
  /* c8 ignore next */
  _Response: _ => _,
  /* c8 ignore next */
  _Security: _ => _
};
/** @internal */
class ApiEndpointImpl {
  id;
  path;
  method;
  request;
  responses;
  security;
  options;
  [TypeId] = variance;
  constructor(id, path, method, request, responses, security, options) {
    this.id = id;
    this.path = path;
    this.method = method;
    this.request = request;
    this.responses = responses;
    this.security = security;
    this.options = options;
  }
  pipe() {
    return Pipeable.pipeArguments(this, arguments);
  }
}
/** @internal */
export const make = (method, id, path, options) => new ApiEndpointImpl(id, path, method, api_request.defaultRequest, [api_response.defaultResponse], Security.unit, {
  ...options
});
/** @internal */
export const setRequest = request => endpoint => new ApiEndpointImpl(getId(endpoint), getPath(endpoint), getMethod(endpoint), request, getResponse(endpoint), getSecurity(endpoint), getOptions(endpoint));
/** @internal */
export const setRequestBody = schema => endpoint => {
  if (getMethod(endpoint) === "GET") {
    throw new Error(`GET ${getPath(endpoint)} (${getId(endpoint)}) cannot have a request body`);
  }
  return setRequest(ApiRequest.setBody(schema)(getRequest(endpoint)))(endpoint);
};
/** @internal */
export const setRequestPath = schema => endpoint => setRequest(ApiRequest.setPath(schema)(getRequest(endpoint)))(endpoint);
/** @internal */
export const setRequestQuery = schema => endpoint => setRequest(ApiRequest.setQuery(schema)(getRequest(endpoint)))(endpoint);
/** @internal */
export const setRequestHeaders = schema => endpoint => setRequest(ApiRequest.setHeaders(schema)(getRequest(endpoint)))(endpoint);
/** @internal */
export const setSecurity = security => endpoint => new ApiEndpointImpl(getId(endpoint), getPath(endpoint), getMethod(endpoint), getRequest(endpoint), getResponse(endpoint), security, getOptions(endpoint));
/** @internal */
export const setOptions = options => endpoint => new ApiEndpointImpl(getId(endpoint), getPath(endpoint), getMethod(endpoint), getRequest(endpoint), getResponse(endpoint), getSecurity(endpoint), {
  ...getOptions(endpoint),
  ...options
});
/** @internal */
export const getId = endpoint => endpoint.id;
/** @internal */
export const getRequest = endpoint => endpoint.request;
/** @internal */
export const getResponse = endpoint => endpoint.responses;
/** @internal */
export const getSecurity = endpoint => endpoint.security;
/** @internal */
export const getMethod = endpoint => endpoint.method;
/** @internal */
export const getPath = endpoint => endpoint.path;
/** @internal */
export const getOptions = endpoint => endpoint.options;
/** @internal */
export const setResponse = response => endpoint => {
  const newResponse = ApiResponse.isApiResponse(response) ? response : ApiResponse.make(response.status, response.body, response.headers);
  return new ApiEndpointImpl(getId(endpoint), getPath(endpoint), getMethod(endpoint), getRequest(endpoint), [newResponse], getSecurity(endpoint), getOptions(endpoint));
};
/** @internal */
export const setResponseStatus = status => endpoint => {
  const responses = getResponse(endpoint);
  if (responses.length !== 1) {
    throw new Error("Expected exactly one response");
  }
  return setResponse(ApiResponse.setStatus(status)(responses[0]))(endpoint);
};
/** @internal */
export const setResponseBody = schema => endpoint => {
  const responses = getResponse(endpoint);
  if (responses.length !== 1) {
    throw new Error("Expected exactly one response");
  }
  return setResponse(ApiResponse.setBody(schema)(responses[0]))(endpoint);
};
/** @internal */
export const setResponseHeaders = schema => endpoint => {
  const responses = getResponse(endpoint);
  if (responses.length !== 1) {
    throw new Error("Expected exactly one response");
  }
  return setResponse(ApiResponse.setHeaders(schema)(responses[0]))(endpoint);
};
export const addResponse = response => endpoint => {
  const responses = getResponse(endpoint);
  const newResponse = ApiResponse.isApiResponse(response) ? response : ApiResponse.make(response.status, response.body, response.headers);
  if (responses.some(r => ApiResponse.getStatus(r) === ApiResponse.getStatus(newResponse))) {
    throw new Error(`Status code ${ApiResponse.getStatus(newResponse)} already exists`);
  }
  return new ApiEndpointImpl(getId(endpoint), getPath(endpoint), getMethod(endpoint), getRequest(endpoint), [...responses, newResponse], getSecurity(endpoint), getOptions(endpoint));
};
export const setResponseRepresentations = representations => endpoint => {
  const responses = getResponse(endpoint);
  if (responses.length !== 1) {
    throw new Error("Expected exactly one response");
  }
  return new ApiEndpointImpl(getId(endpoint), getPath(endpoint), getMethod(endpoint), getRequest(endpoint), [ApiResponse.setRepresentations(representations)(responses[0])], getSecurity(endpoint), getOptions(endpoint));
};
export const isFullResponse = endpoint => {
  const response = getResponse(endpoint);
  const successResponses = response.filter(response => ApiResponse.getStatus(response) < 300);
  const hasMultipleSuccessResponses = successResponses.length > 1;
  const hasHeaders = !ApiSchema.isIgnored(ApiResponse.getHeadersSchema(response[0]));
  return hasMultipleSuccessResponses || hasHeaders;
};
/** @internal */
const getSchemaPropertySignatures = schema => {
  let ast = schema.ast;
  if (ast._tag === "Transformation") {
    ast = ast.from;
  }
  if (ast._tag !== "TypeLiteral") {
    throw new Error(`Response headers must be a type literal schema`);
  }
  return ast.propertySignatures;
};
/** @internal */
const tupleOrder = /*#__PURE__*/Order.tuple(Order.string, Order.boolean);
/** @internal */
const tupleEquivalence = /*#__PURE__*/Equivalence.tuple(Equivalence.string, Equivalence.boolean);
/** @internal */
const arrayOfTupleEquals = /*#__PURE__*/Array.getEquivalence(tupleEquivalence);
/** @internal */
const checkPathPatternMatchesSchema = (id, path, schema) => {
  const fromSchema = pipe(ApiSchema.isIgnored(schema) ? [] : getSchemaPropertySignatures(schema), Array.fromIterable, Array.map(ps => [ps.name, ps.isOptional]), Array.sort(tupleOrder));
  const fromPath = pipe(path.matchAll(/:(\w+)[?]?/g), Array.fromIterable, Array.map(([name]) => {
    if (name.endsWith("?")) {
      return [name.slice(1, -1), true];
    }
    return [name.slice(1), false];
  }), Array.sort(tupleOrder));
  const matched = arrayOfTupleEquals(fromPath, fromSchema);
  if (!matched) {
    const pathNames = fromPath.map(Tuple.getFirst).join(", ");
    const schemaNames = fromSchema.map(Tuple.getFirst).join(", ");
    throw new Error(`Path doesn't match the path schema (endpoint: "${id}").\nfrom path - ${pathNames}\nfrom schema - ${schemaNames}`);
  }
};
/** @internal */
export const validateEndpoint = endpoint => {
  checkPathPatternMatchesSchema(getId(endpoint), getPath(endpoint), ApiRequest.getPathSchema(getRequest(endpoint)));
};
/** @internal */
export const isApiEndpoint = u => Predicate.hasProperty(u, TypeId) && Predicate.isObject(u[TypeId]);
//# sourceMappingURL=api-endpoint.js.map