import * as HttpClientRequest from "@effect/platform/HttpClientRequest";
import * as Effect from "effect/Effect";
import { identity, pipe } from "effect/Function";
import * as Predicate from "effect/Predicate";
import * as Schema from "effect/Schema";
import * as ApiEndpoint from "../ApiEndpoint.js";
import * as ApiRequest from "../ApiRequest.js";
import * as ApiSchema from "../ApiSchema.js";
import * as ClientError from "../ClientError.js";
/** @internal */
const make = encodeRequest => ({
  encodeRequest
});
/** @internal */
export const create = endpoint => {
  const encodeBody = createBodyEncoder(endpoint);
  const encodeQuery = createQueryEncoder(endpoint);
  const encodeHeaders = createHeadersEncoder(endpoint);
  const encodePath = createPathEncoder(endpoint);
  return make(input => Effect.gen(function* (_) {
    const _input = input || {};
    const body = yield* _(encodeBody(_input["body"]));
    const query = yield* _(encodeQuery(_input["query"]));
    const path = yield* _(encodePath(_input["path"]));
    const headers = yield* _(encodeHeaders(_input["headers"]));
    const finalPath = constructPath(path || {}, ApiEndpoint.getPath(endpoint));
    const request = pipe(HttpClientRequest.get(finalPath), HttpClientRequest.setMethod(ApiEndpoint.getMethod(endpoint)), body === undefined ? identity : body instanceof FormData ? HttpClientRequest.bodyFormData(body) : HttpClientRequest.bodyUnsafeJson(body), query ? HttpClientRequest.setUrlParams(query) : identity, headers ? HttpClientRequest.setHeaders(headers) : identity);
    return request;
  }));
};
/** @internal */
const ignoredSchemaEncoder = name => input => {
  if (input !== undefined) {
    return Effect.dieMessage(`Unexpected ${name} provided, got ${JSON.stringify(input)}`);
  }
  return Effect.succeed(undefined);
};
/** @internal */
const createBodyEncoder = endpoint => {
  const schema = ApiRequest.getBodySchema(ApiEndpoint.getRequest(endpoint));
  if (ApiSchema.isIgnored(schema)) {
    return ignoredSchemaEncoder("body");
  }
  const encode = Schema.encode(schema);
  return body => {
    return encode(body).pipe(Effect.mapError(error => ClientError.makeClientSide(error, `Failed to encode body. ${error.message}`)));
  };
};
/** @internal */
const isRecordOrUndefined = i => Predicate.isRecord(i) || Predicate.isUndefined(i);
/** @internal */
const createQueryEncoder = endpoint => {
  const schema = ApiRequest.getQuerySchema(ApiEndpoint.getRequest(endpoint));
  if (ApiSchema.isIgnored(schema)) {
    return ignoredSchemaEncoder("query");
  }
  const encode = Schema.encode(schema);
  return query => {
    return encode(query).pipe(Effect.mapError(error => ClientError.makeClientSide(error, `Failed to encode query parameters. ${error.message}`)));
  };
};
/** @internal */
const createHeadersEncoder = endpoint => {
  const schema = ApiRequest.getHeadersSchema(ApiEndpoint.getRequest(endpoint));
  const encode = ApiSchema.isIgnored(schema) ? undefined : Schema.encode(schema);
  return headers => {
    if (!isRecordOrUndefined(headers)) {
      return Effect.dieMessage("Headers must be a record");
    }
    return (encode ?? Effect.succeed)(headers).pipe(Effect.mapError(error => ClientError.makeClientSide(error, `Failed to encode headers. ${error.message}`)));
  };
};
/** @internal */
const createPathEncoder = endpoint => {
  const schema = ApiRequest.getPathSchema(ApiEndpoint.getRequest(endpoint));
  if (ApiSchema.isIgnored(schema)) {
    return ignoredSchemaEncoder("path");
  }
  const encode = Schema.encode(schema);
  return params => {
    return encode(params).pipe(Effect.mapError(error => ClientError.makeClientSide(error, `Failed to encode path parmeters. ${error.message}.`)));
  };
};
/** @internal */
const constructPath = (params, path) => {
  return Object.entries(params ?? {}).reduce((path, [key, value]) => path.replace(new RegExp(`(:${key})(\\?)?`), value), path).replace(/\/:(\w+)(\?)?/, "");
};
//# sourceMappingURL=clientRequestEncoder.js.map