import * as Headers from "@effect/platform/Headers";
import * as HttpServerRequest from "@effect/platform/HttpServerRequest";
import * as Effect from "effect/Effect";
import * as Either from "effect/Either";
import * as Encoding from "effect/Encoding";
import { dual, pipe } from "effect/Function";
import * as Option from "effect/Option";
import * as Pipeable from "effect/Pipeable";
import * as Schema from "effect/Schema";
import * as Unify from "effect/Unify";
import * as HttpError from "../HttpError.js";
/** @internal */
export const TypeId = /*#__PURE__*/Symbol.for("effect-http/Security/TypeId");
/** @internal */
export const variance = {
  /* c8 ignore next */
  _A: _ => _,
  /* c8 ignore next */
  _E: _ => _,
  /* c8 ignore next */
  _R: _ => _
};
/** @internal */
export class SecurityImpl {
  openapi;
  parser;
  [TypeId] = variance;
  constructor(openapi, parser) {
    this.openapi = openapi;
    this.parser = parser;
  }
  pipe() {
    return Pipeable.pipeArguments(this, arguments);
  }
}
/** @internal */
export const make = (parser, openapi) => new SecurityImpl(openapi ?? {}, parser);
/** @internal */
export const handleRequest = security => security.parser;
/** @internal */
export const getOpenApi = security => security.openapi;
/** @internal */
export const mapHandler = /*#__PURE__*/dual(2, (self, f) => make(f(handleRequest(self)), getOpenApi(self)));
/** @internal */
export const and = /*#__PURE__*/dual(2, (self, that) => make(Effect.all([handleRequest(self), handleRequest(that)]), {
  ...getOpenApi(self),
  ...getOpenApi(that)
}));
/** @internal */
export const or = /*#__PURE__*/dual(2, (self, that) => make(Effect.orElse(handleRequest(self), () => handleRequest(that)), {
  ...getOpenApi(self),
  ...getOpenApi(that)
}));
/** @internal */
export const mapEffect = /*#__PURE__*/dual(2, (self, f) => mapHandler(self, Effect.flatMap(f)));
/** @internal */
export const mapSchema = /*#__PURE__*/dual(2, (self, schema) => {
  const parse = Schema.decode(schema);
  return make(Effect.flatMap(handleRequest(self), token => pipe(parse(token), Effect.mapError(error => HttpError.unauthorized(`Security parsing error, ${error}`)))), getOpenApi(self));
});
/** @internal */
export const as = /*#__PURE__*/dual(2, (self, value) => map(self, () => value));
/** @internal */
export const map = /*#__PURE__*/dual(2, (self, f) => make(Effect.map(handleRequest(self), f), getOpenApi(self)));
export const asSome = self => map(self, Option.some);
/** @internal */
const getAuthorizationHeader = /*#__PURE__*/pipe(HttpServerRequest.HttpServerRequest, /*#__PURE__*/Effect.flatMap(request => Headers.get(request.headers, "authorization")), /*#__PURE__*/Effect.mapError(() => HttpError.unauthorized("No authorization header")));
/** @internal */
const parseAuthorizationHeader = scheme => Unify.unify(value => {
  const split = value.split(" ");
  if (split.length !== 2) {
    return Effect.fail(HttpError.unauthorized("Invalid authorization header"));
  }
  if (split[0].toLowerCase() !== scheme) {
    return Effect.fail(HttpError.unauthorized(`Expected ${scheme.toUpperCase()} authorization`));
  }
  return Effect.succeed(split[1]);
});
/** @internal */
export const bearer = options => {
  const description = options?.description;
  const bearerFormat = options?.bearerFormat;
  const openApi = {
    [options?.name ?? "bearer"]: {
      type: "http",
      ...(description && {
        description
      }),
      ...(bearerFormat && {
        bearerFormat
      }),
      scheme: "bearer"
    }
  };
  return make(Effect.flatMap(getAuthorizationHeader, parseAuthorizationHeader("bearer")), openApi);
};
/** @internal */
const parseBasicAuthCredentials = /*#__PURE__*/Unify.unify(value => pipe(Encoding.decodeBase64String(value), Either.mapLeft(() => HttpError.unauthorized("Invalid base64-encoded basic authorization")), Either.flatMap(value => {
  const split = value.split(":");
  if (split.length !== 2) {
    return Either.left(HttpError.unauthorized(`Invalid basic authorization header, expected "user:password".`));
  }
  return Either.right({
    user: split[0],
    pass: split[1]
  });
})));
/** @internal */
export const basic = options => {
  const description = options?.description;
  const openApi = {
    [options?.name ?? "basic"]: {
      type: "http",
      ...(description && {
        description
      }),
      scheme: "basic"
    }
  };
  return make(pipe(getAuthorizationHeader, Effect.flatMap(parseAuthorizationHeader("basic")), Effect.flatMap(parseBasicAuthCredentials)), openApi);
};
/** @internal */
export const unit = /*#__PURE__*/make(Effect.void);
/** @internal */
export const never = /*#__PURE__*/make(/*#__PURE__*/Effect.fail(/*#__PURE__*/HttpError.unauthorized("No security")));
/** @internal */
export const apiKey = options => {
  const schema = Schema.Struct({
    [options.key]: Schema.String
  });
  const parse = pipe(Unify.unify(options.in === "query" ? HttpServerRequest.schemaSearchParams(schema) : HttpServerRequest.schemaHeaders(schema)), Effect.map(obj => obj[options.key]), Effect.mapError(() => HttpError.unauthorized(`Expected ${options.key} (${options.in}) api key`)));
  return make(parse, {
    [options.name ?? "apiKey"]: {
      name: options.key,
      type: "apiKey",
      in: options.in,
      ...(options.description && {
        description: options.description
      })
    }
  });
};
//# sourceMappingURL=security.js.map