import * as HttpRouter from "@effect/platform/HttpRouter";
import * as HttpServerRequest from "@effect/platform/HttpServerRequest";
import * as Effect from "effect/Effect";
import * as Schema from "effect/Schema";
import * as ApiEndpoint from "../ApiEndpoint.js";
import * as ApiRequest from "../ApiRequest.js";
import * as ApiSchema from "../ApiSchema.js";
import * as HttpError from "../HttpError.js";
import * as Security from "../Security.js";
/** @internal */
const createError = (location, message) => HttpError.make(400, {
  error: "Request validation error",
  location,
  message
});
/** @internal */
const make = parseRequest => ({
  parseRequest
});
/** @internal */
export const create = (endpoint, parseOptions) => make(Effect.all({
  body: parseBody(endpoint, parseOptions),
  query: parseQuery(endpoint, parseOptions),
  path: parsePath(endpoint, parseOptions),
  headers: parseHeaders(endpoint, parseOptions),
  security: parseSecurity(endpoint)
}));
/** @internal */
const parseBody = (endpoint, parseOptions) => {
  const schema = ApiRequest.getBodySchema(ApiEndpoint.getRequest(endpoint));
  if (ApiSchema.isIgnored(schema)) {
    return Effect.succeed(undefined);
  }
  const parse = Schema.decodeUnknown(schema);
  if (schema === ApiSchema.FormData) {
    // TODO
    return Effect.succeed(undefined);
  }
  return HttpServerRequest.HttpServerRequest.pipe(Effect.flatMap(request => request.json), Effect.mapError(error => {
    if (error.reason === "Transport") {
      return createError("body", "Unexpect request JSON body error");
    }
    return createError("body", "Invalid JSON");
  }), Effect.flatMap(request => parse(request, parseOptions).pipe(Effect.mapError(error => createError("body", error.message)))));
};
/** @internal */
const parseQuery = (endpoint, parseOptions) => {
  const schema = ApiRequest.getQuerySchema(ApiEndpoint.getRequest(endpoint));
  if (ApiSchema.isIgnored(schema)) {
    return Effect.succeed(undefined);
  }
  return Effect.mapError(HttpServerRequest.schemaSearchParams(schema, parseOptions), error => createError("query", error.message));
};
/** @internal */
const parseHeaders = (endpoint, parseOptions) => {
  const schema = ApiRequest.getHeadersSchema(ApiEndpoint.getRequest(endpoint));
  if (ApiSchema.isIgnored(schema)) {
    return Effect.succeed(undefined);
  }
  const parse = Schema.decodeUnknown(schema);
  return HttpServerRequest.HttpServerRequest.pipe(Effect.flatMap(request => parse(request.headers, parseOptions)), Effect.mapError(error => createError("headers", error.message)));
};
/** @internal */
const parseSecurity = endpoint => {
  const security = ApiEndpoint.getSecurity(endpoint);
  return Security.handleRequest(security);
};
/** @internal */
const parsePath = (endpoint, parseOptions) => {
  const schema = ApiRequest.getPathSchema(ApiEndpoint.getRequest(endpoint));
  if (ApiSchema.isIgnored(schema)) {
    return Effect.succeed(undefined);
  }
  const parse = Schema.decodeUnknown(schema);
  return HttpRouter.RouteContext.pipe(Effect.flatMap(ctx => parse(ctx.params, parseOptions)), Effect.mapError(error => createError("path", error.message)));
};
//# sourceMappingURL=serverRequestParser.js.map